from collections.abc import Mapping
from typing import Any

import attr

from metric_config_parser.analysis import AnalysisSpec
from metric_config_parser.monitoring import MonitoringSpec

from .alert import AlertsSpec
from .data_source import DataSourcesSpec
from .dimension import DimensionsSpec
from .metric import MetricsSpec
from .parameter import ParameterSpec
from .segment import SegmentsSpec
from .util import converter


@attr.s(auto_attribs=True)
class DefinitionSpec:
    """A representation of a collection of definitions."""

    metrics: MetricsSpec = attr.Factory(MetricsSpec)
    data_sources: DataSourcesSpec = attr.Factory(DataSourcesSpec)
    segments: SegmentsSpec = attr.Factory(SegmentsSpec)
    dimensions: DimensionsSpec = attr.Factory(DimensionsSpec)
    alerts: AlertsSpec = attr.Factory(AlertsSpec)
    parameters: ParameterSpec = attr.Factory(ParameterSpec)

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> "DefinitionSpec":
        return converter.structure(d, cls)

    def merge(self, other: "DefinitionSpecSub"):
        self.metrics.merge(other.metrics)
        self.data_sources.merge(other.data_sources)
        if isinstance(other, DefinitionSpec | AnalysisSpec):
            self.segments.merge(other.segments)
        if isinstance(other, DefinitionSpec | MonitoringSpec):
            self.dimensions.merge(other.dimensions)
        if isinstance(other, DefinitionSpec | MonitoringSpec):
            self.alerts.merge(other.alerts)


DefinitionSpecSub = AnalysisSpec | MonitoringSpec | DefinitionSpec



"""CLI."""

import os
import sys
from collections.abc import Iterable
from pathlib import Path

import click

from metric_config_parser.config import (
    DEFINITIONS_DIR,
    ConfigCollection,
    DefinitionConfig,
    Outcome,
    entity_from_path,
)
from metric_config_parser.function import FunctionsSpec

METRIC_HUB_REPO = "https://github.com/mozilla/metric-hub"


@click.group()
def cli():
    """Initialize CLI."""
    pass


@cli.command("validate")
@click.argument("path", type=click.Path(exists=True), nargs=-1)
@click.option(
    "--config_repos",
    "--config-repos",
    help="URLs to public repos with configs",
    multiple=True,
    default=[METRIC_HUB_REPO],
)
@click.option(
    "--private_config_repos",
    "--private-config-repos",
    help="URLs to private repos with configs",
    multiple=True,
)
def validate(path: Iterable[os.PathLike], config_repos, private_config_repos):
    """Validate config files."""
    dirty = False
    config_collection = ConfigCollection.from_github_repos(config_repos).from_github_repos(
        private_config_repos, is_private=True
    )

    # get updated definition files
    for config_file in path:
        config_file = Path(config_file)
        if not config_file.is_file():
            continue
        if ".example" in config_file.suffixes:
            print(f"Skipping example config {config_file}")
            continue

        if config_file.parent.name == DEFINITIONS_DIR:
            entity = entity_from_path(config_file)
            try:
                if isinstance(entity, Outcome):
                    entity.validate(config_collection)
                elif not isinstance(entity, FunctionsSpec):
                    entity.validate(config_collection, None)  # type: ignore
            except Exception as e:
                dirty = True
                print(e)
            else:
                print(f"{config_file} OK")

            if isinstance(entity, DefinitionConfig):
                config_collection.definitions.append(entity)

    for config_file in path:
        config_file = Path(config_file)
        if config_file.parent.name == DEFINITIONS_DIR:
            continue
        if not config_file.is_file():
            continue
        if ".example" in config_file.suffixes:
            print(f"Skipping example config {config_file}")
            continue
        print(f"Evaluating {config_file}...")
        entity = entity_from_path(config_file)
        try:
            if not isinstance(entity, FunctionsSpec) and not isinstance(entity, Outcome):
                entity.validate(config_collection, None)  # type: ignore
        except Exception as e:
            dirty = True
            print(e)
        else:
            print(f"{config_file} OK")

    sys.exit(1 if dirty else 0)



import datetime
import re
import shutil
from pathlib import Path
from textwrap import dedent

import pytest
import pytz
import toml
from cattrs.errors import ClassValidationError
from git import Repo
from git.exc import InvalidGitRepositoryError

from metric_config_parser import AnalysisUnit
from metric_config_parser.analysis import AnalysisSpec
from metric_config_parser.config import (
    Config,
    ConfigCollection,
    DefaultConfig,
    DefinitionConfig,
    LocalConfigCollection,
    Outcome,
)
from metric_config_parser.data_source import DataSourceJoinRelationship
from metric_config_parser.errors import DefinitionNotFound
from metric_config_parser.metric import MetricLevel
from metric_config_parser.outcome import OutcomeSpec

TEST_DIR = Path(__file__).parent


class TestConfigIntegration:
    config_str = dedent(
        """
        [metrics]
        weekly = ["active_hours"]

        [metrics.active_hours.statistics.bootstrap_mean]
        """
    )
    spec = AnalysisSpec.from_dict(toml.loads(config_str))

    def test_old_config(self):
        config = Config(
            slug="new_table",
            spec=self.spec,
            last_modified=pytz.UTC.localize(
                datetime.datetime.utcnow() - datetime.timedelta(days=1)
            ),
        )

        config_collection = ConfigCollection([config])

        assert config_collection.spec_for_experiment("new_table") is not None
        assert config_collection.spec_for_outcome("test", "foo") is None
        assert config_collection.get_platform_defaults("desktop") is None
        assert config_collection.get_segment_data_source_definition("foo", "test") is None

    def test_definition_config(self):
        config_str = dedent(
            """
            [metrics.retained]
            select_expression = "COALESCE(COUNT(document_id), 0) > 0"
            data_source = "baseline"

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )

        assert definition

    def test_valid_config_validates(self, experiments):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )

        config_collection = ConfigCollection(
            configs=[extern], outcomes=[], defaults=[], definitions=[definition]
        )
        extern.validate(config_collection, experiments[0])

    def test_busted_config_fails(self, experiments):
        config = dedent(
            """\
            [metrics]
            weekly = ["bogus_metric"]

            [metrics.bogus_metric]
            select_expression = "SUM(fake_column)"
            data_source = "clients_daily"
            statistics = { bootstrap_mean = {} }
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(config))
        extern = Config(
            slug="bad_experiment",
            spec=spec,
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection([extern])
        with pytest.raises(DefinitionNotFound):
            extern.validate(config_collection, experiments[0])

    def test_valid_outcome_validates(self):
        config = dedent(
            """\
            friendly_name = "Fred"
            description = "Just your average paleolithic dad."

            [metrics.rocks_mined]
            select_expression = "COALESCE(SUM(pings_aggregated_by_this_row), 0)"
            data_source = "clients_daily"
            statistics = { bootstrap_mean = {} }
            friendly_name = "Rocks mined"
            description = "Number of rocks mined at the quarry"

            [data_sources.clients_daily]
            from_expression = "1"
            """
        )
        spec = OutcomeSpec.from_dict(toml.loads(config))
        extern = Outcome(
            slug="good_outcome",
            spec=spec,
            platform="firefox_desktop",
            commit_hash="0000000",
        )
        extern.validate(configs=ConfigCollection())

    def test_busted_outcome_fails(self):
        config = dedent(
            """\
            friendly_name = "Fred"
            description = "Just your average paleolithic dad."

            [metrics.rocks_mined]
            select_expression = "COALESCE(SUM(fake_column_whoop_whoop), 0)"
            data_source = "clients_daily"
            statistics = { bootstrap_mean = {} }
            friendly_name = "Rocks mined"
            description = "Number of rocks mined at the quarry"
            """
        )
        spec = OutcomeSpec.from_dict(toml.loads(config))
        extern = Outcome(
            slug="bogus_outcome",
            spec=spec,
            platform="firefox_desktop",
            commit_hash="0000000",
        )
        with pytest.raises(DefinitionNotFound):
            extern.validate(configs=ConfigCollection())

    def test_valid_default_config_validates(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        extern = DefaultConfig(
            slug="firefox_desktop",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        extern.validate(configs=ConfigCollection(definitions=[definition]))

    def test_busted_default_config_fails(self):
        config = dedent(
            """\
            [metrics]
            weekly = ["bogus_metric"]

            [metrics.bogus_metric]
            select_expression = "SUM(fake_column)"
            data_source = "clients_daily"
            statistics = { bootstrap_mean = {} }
            """
        )
        spec = AnalysisSpec.from_dict(toml.loads(config))
        extern = DefaultConfig(
            slug="firefox_desktop",
            spec=spec,
            last_modified=datetime.datetime.now(),
        )
        with pytest.raises(DefinitionNotFound):
            extern.validate(configs=ConfigCollection())

    def test_merge_config_collection(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_1 = ConfigCollection(
            configs=[extern], outcomes=[], defaults=[], definitions=[definition]
        )

        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment_2",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_2 = ConfigCollection(
            configs=[extern], outcomes=[], defaults=[], definitions=[definition]
        )

        config_collection_1.merge(config_collection_2)
        assert config_collection_1.configs[0].slug == "cool_experiment"
        assert config_collection_1.configs[1].slug == "cool_experiment_2"

    def test_merge_config_collection_override(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        config_collection_1 = ConfigCollection(
            configs=[extern], outcomes=[], defaults=[], definitions=[]
        )

        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "4"
            data_source = "baseline"

            [metrics.active_hours.statistics.bootstrap_mean]

            [metrics.unenroll]
            select_expression = "3"
            data_source = "baseline"

            [metrics.unenroll.statistics.bootstrap_mean]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )
        extern = Config(
            slug="cool_experiment",
            spec=self.spec,
            last_modified=datetime.datetime.now(),
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_2 = ConfigCollection(
            configs=[extern], outcomes=[], defaults=[], definitions=[definition]
        )

        assert len(config_collection_1.definitions) == 0
        config_collection_1.merge(config_collection_2)

        assert len(config_collection_1.configs) == 1
        assert config_collection_1.configs[0].slug == "cool_experiment"

        assert (
            next(
                m
                for slug, m in config_collection_1.definitions[0].spec.metrics.definitions.items()
                if slug == "active_hours"
            ).select_expression
            == "4"
        )
        assert (
            next(
                m
                for slug, m in config_collection_1.definitions[0].spec.metrics.definitions.items()
                if slug == "unenroll"
            ).select_expression
            == "3"
        )

    def test_config_collection_from_subdir(self, local_tmp_repo):
        config_collection = ConfigCollection.from_github_repo(
            local_tmp_repo, path="metrics/jetstream"
        )
        assert len(config_collection.configs) > 0

    def test_configs_from_private_repo(self, local_tmp_repo):
        config_collection = ConfigCollection.from_github_repo(
            local_tmp_repo, path="metrics/jetstream", is_private=True
        )
        assert config_collection is not None
        assert config_collection.configs[0].spec.experiment.is_private

    def test_config_from_subdir(self, local_tmp_repo):
        nested_path = Path(local_tmp_repo) / "metrics" / "jetstream"
        config_collection = ConfigCollection.from_github_repo(nested_path)
        assert config_collection is not None

    def test_config_from_subdir_too_deep(self, local_tmp_repo):
        nested_path = Path(local_tmp_repo) / "metrics" / "jetstream"

        with pytest.raises(InvalidGitRepositoryError):
            ConfigCollection.from_github_repo(nested_path, depth=1)

    def test_as_of_broken_commit(self, tmp_path):
        r = Repo.init(tmp_path)
        r.config_writer().set_value("user", "name", "test").release()
        r.config_writer().set_value("user", "email", "test@example.com").release()
        r.config_writer().set_value("commit", "gpgsign", "false").release()

        # check in broken file
        broken_config = dedent(
            """
            friendly_name = "Performance outcomes"
            description = "Outcomes related to performance"

            [metrics]
            weekly = ["speed"]
            overall = ["speed"]

            [metrics.speed]
            data_source = "main"
            select_expression = "1"

            [metrics.speed.statistics.bootstrap_mean]
            """
        )
        outcome_path = tmp_path / "jetstream" / "outcomes" / "firefox_desktop"
        outcome_path.mkdir(parents=True, exist_ok=True)
        (outcome_path / "performance.toml").write_text(broken_config)
        r.git.add(".")
        r.git.commit("-m", "commit", "--date", "Mon 20 Aug 2020 20:19:19 UTC")

        with pytest.raises(
            AttributeError, match=re.escape("'list' object has no attribute 'items'")
        ):
            ConfigCollection.from_github_repo(tmp_path / "jetstream")

        with pytest.raises(
            AttributeError, match=re.escape("'list' object has no attribute 'items'")
        ):
            ConfigCollection.from_github_repo(tmp_path / "jetstream").as_of(
                pytz.UTC.localize(datetime.datetime(2023, 5, 21))
            )

        # check in valid files
        shutil.copytree(TEST_DIR / "data", tmp_path, dirs_exist_ok=True)
        r.git.add(".")
        r.git.commit("-m", "commit", "--date", "Mon 25 Aug 2020 20:00:19 UTC")

        configs = ConfigCollection.from_github_repo(tmp_path / "jetstream")
        assert configs.outcomes is not None

        configs = configs.as_of(pytz.UTC.localize(datetime.datetime(2023, 5, 21)))
        assert configs.outcomes is not None

    def test_metric_level(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"
            owner = "me@example.com"
            category = "test"
            level = "silver"

            [metrics.other_metric]
            select_expression = "1"
            data_source = "baseline"
            owner = ["me@example.com", "you@example.com"]
            category = "test"
            level = "gold"

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        metric_definition = config_collection.get_metric_definition(
            "active_hours", "firefox_desktop"
        )

        assert metric_definition.level == MetricLevel.SILVER
        assert metric_definition.owner == "me@example.com"
        assert metric_definition.category == "test"

        metric_definition = config_collection.get_metric_definition(
            "other_metric", "firefox_desktop"
        )

        assert metric_definition.level == MetricLevel.GOLD
        assert metric_definition.owner == ["me@example.com", "you@example.com"]
        assert metric_definition.category == "test"

    def test_invalid_metric_level(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"
            level = "invalid"

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )

        with pytest.raises(ClassValidationError):
            DefinitionConfig(
                slug="firefox_desktop",
                platform="firefox_desktop",
                spec=AnalysisSpec.from_dict(toml.loads(config_str)),
                last_modified=datetime.datetime.now(),
            )

    def test_valid_analysis_units(self):
        config_str = dedent(
            """
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"
            owner = "me@example.com"
            category = "test"
            analysis_units = ["profile_group_id"]

            [metrics.other_metric]
            select_expression = "1"
            data_source = "baseline"
            owner = ["me@example.com", "you@example.com"]
            category = "test"
            analysis_units = ["client_id", "profile_group_id"]

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            analysis_units = ["client_id", "profile_group_id"]
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        metric_definition = config_collection.get_metric_definition(
            "active_hours", "firefox_desktop"
        )
        assert metric_definition.analysis_units == [AnalysisUnit.PROFILE_GROUP]

        metric_definition = config_collection.get_metric_definition(
            "other_metric", "firefox_desktop"
        )
        assert metric_definition.analysis_units == [
            AnalysisUnit.CLIENT,
            AnalysisUnit.PROFILE_GROUP,
        ]

        data_source_definition = config_collection.get_data_source_definition(
            "baseline", "firefox_desktop"
        )
        assert data_source_definition.analysis_units == [
            AnalysisUnit.CLIENT,
            AnalysisUnit.PROFILE_GROUP,
        ]

    @pytest.mark.parametrize(
        ("metric_units", "ds_units"),
        [
            ("analysis_units = 'client_id'", ""),
            ("analysis_units = ['invalid']", ""),
            ("", "analysis_units = 'client_id'"),
            ("", "analysis_units = ['invalid']"),
        ],
    )
    def test_invalid_analysis_units(self, metric_units, ds_units):
        config_str = dedent(
            f"""
            [metrics.active_hours]
            select_expression = "1"
            data_source = "baseline"
            {metric_units}

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            {ds_units}
            """
        )
        with pytest.raises(ClassValidationError):
            DefinitionConfig(
                slug="firefox_desktop",
                platform="firefox_desktop",
                spec=AnalysisSpec.from_dict(toml.loads(config_str)),
                last_modified=datetime.datetime.now(),
            )

    def test_data_source_joins(self):
        config_str = dedent(
            """
            [data_sources.events]
            from_expression = "mozdata.telemetry.events"
            experiments_column_type = "simple"

            [data_sources.metrics]
            from_expression = "mozdata.telemetry.metrics"
            experiments_column_type = "simple"

            [data_sources.baseline]
            from_expression = "mozdata.telemetry.baseline"
            experiments_column_type = "simple"

            [data_sources.baseline.joins.metrics]
            on_expression = "metrics.client_id = baseline.client_id"
            relationship = "many_to_one"

            [data_sources.baseline.joins.events]
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        data_source = config_collection.get_data_source_definition(
            "baseline", "firefox_desktop"
        ).resolve(definition.spec, None, config_collection)

        assert len(data_source.joins) == 2
        assert data_source.joins[0].data_source.name == "metrics"
        assert data_source.joins[0].on_expression == "metrics.client_id = baseline.client_id"
        assert data_source.joins[0].relationship == DataSourceJoinRelationship.MANY_TO_ONE

        assert data_source.joins[1].data_source.name == "events"

    def test_data_source_joins_invalid(self):
        config_str = dedent(
            """
            [data_sources.baseline]
            from_expression = "mozdata.telemetry.baseline"
            experiments_column_type = "simple"

            [data_sources.baseline.joins.non_existing]
            relationship = "many_to_one"
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        with pytest.raises(
            AttributeError, match=re.escape("'NoneType' object has no attribute 'app_name'")
        ):
            config_collection.get_data_source_definition("baseline", "firefox_desktop").resolve(
                definition.spec, None, config_collection
            )

    def test_data_source_joins_circular_dependency(self):
        config_str = dedent(
            """
            [data_sources.baseline]
            from_expression = "mozdata.telemetry.baseline"
            experiments_column_type = "simple"

            [data_sources.metrics]
            from_expression = "mozdata.telemetry.metrics"
            experiments_column_type = "simple"

            [data_sources.baseline.joins.metrics]

            [data_sources.metrics.joins.baseline]
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        with pytest.raises(RecursionError, match="maximum recursion depth exceeded"):
            config_collection.get_data_source_definition("baseline", "firefox_desktop").resolve(
                definition.spec, None, config_collection
            )

    def test_invalid_wildcard_in_data_source(self):
        # needs to be [data_sources.'baseline_*']
        config_str = dedent(
            """
            [data_sources.baseline_*]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )

        with pytest.raises(
            ValueError,
            match=re.escape(
                "Invalid group name 'baseline_*'. Try quoting it. (line 2 column 1 char 1)"
            ),
        ):
            DefinitionConfig(
                slug="firefox_desktop",
                platform="firefox_desktop",
                spec=AnalysisSpec.from_dict(toml.loads(config_str)),
                last_modified=datetime.datetime.now(),
            )

    def test_merge_with_wildcards(self):
        config_str = dedent(
            """
            [metrics.test_metric]
            select_expression = 1
            category = "test"

            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_1 = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        config_str = dedent(
            """
            [data_sources.'*']
            experiments_column_type = "none"

            [metrics.'test_*']
            data_source = "baseline"
            category = "test_overwrite_first"

            [metrics.'test_*'.statistics.bootstrap_mean]

            [metrics.'test_me*']
            category = "test_overwrite_second"
            """
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_2 = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        assert (
            config_collection_1.get_metric_definition("test_metric", "firefox_desktop").data_source
            is None
        )
        assert (
            config_collection_1.get_data_source_definition(
                "baseline", "firefox_desktop"
            ).experiments_column_type
            == "simple"
        )

        config_collection_1.merge(config_collection_2)
        assert (
            config_collection_1.get_data_source_definition("baseline", "firefox_desktop").name
            == "baseline"
        )
        assert (
            config_collection_1.get_data_source_definition(
                "baseline", "firefox_desktop"
            ).experiments_column_type
            == "none"
        )
        assert config_collection_1.get_data_source_definition("*", "firefox_desktop") is None
        assert (
            config_collection_1.get_metric_definition(
                "test_metric", "firefox_desktop"
            ).data_source.name
            == "baseline"
        )
        assert (
            "bootstrap_mean"
            in config_collection_1.get_metric_definition(
                "test_metric", "firefox_desktop"
            ).statistics
        )
        assert (
            config_collection_1.get_metric_definition("test_metric", "firefox_desktop").category
            == "test_overwrite_second"
        )

    def test_merge_with_wildcards_invalid(self):
        config_str = dedent(
            """
            [data_sources.baseline]
            from_expression = "mozdata.search.baseline"
            experiments_column_type = "simple"
            """
        )

        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_1 = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        config_str = dedent(
            """
            [data_sources.'invalid_*']
            experiments_column_type = "none"
            """
        )
        definition = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str)),
            last_modified=datetime.datetime.now(),
        )
        config_collection_2 = ConfigCollection(
            configs=[], outcomes=[], defaults=[], definitions=[definition]
        )

        config_collection_1.merge(config_collection_2)
        assert (
            config_collection_1.get_data_source_definition("baseline", "firefox_desktop").name
            == "baseline"
        )
        assert (
            config_collection_1.get_data_source_definition(
                "baseline", "firefox_desktop"
            ).experiments_column_type
            == "simple"
        )
        assert (
            config_collection_1.get_data_source_definition("invalid_*", "firefox_desktop") is None
        )

    def test_get_segments_for_app_with_toml(self):
        config_str_desktop = """
        [segments.my_cool_segment]
        data_source = "my_cool_data_source"
        select_expression = "{{agg_any('1')}}"

        [segments.another_segment]
        data_source = "another_data_source"
        select_expression = "{{agg_sum('2')}}"

        [segments.data_sources.my_cool_data_source]
        from_expression = "max(attr_source is not null)"

        [segments.data_sources.another_data_source]
        from_expression = "max(attr_source is null)"
        """

        config_str_fenix = """
        [segments.fenix_segment]
        data_source = "fenix_data_source"
        select_expression = "{{agg_count('3')}}"

        [segments.data_sources.fenix_data_source]
        from_expression = "min(attr_source is not null)"
        """

        definition_desktop = DefinitionConfig(
            slug="firefox_desktop",
            platform="firefox_desktop",
            spec=AnalysisSpec.from_dict(toml.loads(config_str_desktop)),
            last_modified=datetime.datetime.now(),
        )

        definition_fenix = DefinitionConfig(
            slug="fenix",
            platform="fenix",
            spec=AnalysisSpec.from_dict(toml.loads(config_str_fenix)),
            last_modified=datetime.datetime.now(),
        )

        config_collection = ConfigCollection(
            configs=[],
            outcomes=[],
            defaults=[],
            definitions=[definition_desktop, definition_fenix],
        )

        segments_desktop = config_collection.get_segments_for_app("firefox_desktop")
        assert len(segments_desktop) == 2

        segment_slugs_desktop = [seg.name for seg in segments_desktop]
        assert "my_cool_segment" in segment_slugs_desktop
        assert "another_segment" in segment_slugs_desktop

        segments_fenix = config_collection.get_segments_for_app("fenix")
        assert len(segments_fenix) == 1

        segment_slugs_fenix = [seg.name for seg in segments_fenix]
        assert "fenix_segment" in segment_slugs_fenix

    def test_local_config_collection_from_local_path(self):
        config_collection = LocalConfigCollection.from_local_path(TEST_DIR / "data")
        segments = [s.name for s in config_collection.get_segments_for_app("firefox_desktop")]
        assert "regular_users_v3" in segments
        assert len(segments) == 1

        outcomes = [o.slug for o in config_collection.outcomes]
        assert len(outcomes) == 0

        test_metric = config_collection.get_metric_definition("active_hours", "firefox_desktop")
        assert test_metric.name == "active_hours"
        assert test_metric.select_expression == '{{agg_sum("active_hours_sum")}}'
        assert test_metric.data_source.name == "clients_daily"

        cc_jetstream = LocalConfigCollection.from_local_path(TEST_DIR / "data" / "jetstream")
        config_collection.merge(cc_jetstream)

        outcomes = [o.slug for o in config_collection.outcomes]
        assert len(outcomes) == 4



from typing import TYPE_CHECKING

import attr

from metric_config_parser.metric import MetricReference
from metric_config_parser.util import converter

if TYPE_CHECKING:
    from metric_config_parser.config import ConfigCollection
    from metric_config_parser.monitoring import MonitoringSpec
    from metric_config_parser.project import ProjectConfiguration


@attr.s(auto_attribs=True, kw_only=True)
class MetricGroup:
    """Represents a set of metrics that are related and should be displayed together."""

    name: str
    description: str | None = None
    friendly_name: str | None = None
    metrics: list[MetricReference] = attr.Factory(list)


@attr.s(auto_attribs=True)
class MetricGroupDefinition:
    """Describes the interface for defining a metric group in configuration."""

    name: str  # implicit in configuration
    friendly_name: str | None = None
    description: str | None = None
    metrics: list[MetricReference] = attr.Factory(list)

    def resolve(
        self,
        spec: "MonitoringSpec",
        _conf: "ProjectConfiguration",
        _configs: "ConfigCollection",
    ) -> MetricGroup:
        """Create and return a `MetricGroup` from the definition."""
        for metric_ref in self.metrics:
            if metric_ref not in spec.project.metrics:
                raise ValueError(
                    f"Metric {metric_ref} is part of metric group {self.name} "
                    + "but not referenced as project metric"
                )

        return MetricGroup(
            name=self.name,
            metrics=self.metrics,
            friendly_name=self.friendly_name,
            description=self.description,
        )

    def merge(self, other: "MetricGroupDefinition"):
        """Merge with another metric group definition."""
        for key in attr.fields_dict(type(self)):
            setattr(self, key, getattr(other, key) or getattr(self, key))


@attr.s(auto_attribs=True)
class MetricGroupsSpec:
    """Describes the interface for defining custom dimensions."""

    definitions: dict[str, MetricGroupDefinition] = attr.Factory(dict)

    @classmethod
    def from_dict(cls, d: dict) -> "MetricGroupsSpec":
        """Create a `MetricGroupsSpec` from a dictionary."""
        d = {k.lower(): v for k, v in d.items()}

        definitions = {
            k: converter.structure({"name": k, **v}, MetricGroupDefinition) for k, v in d.items()
        }
        return cls(definitions=definitions)

    def merge(self, other: "MetricGroupsSpec"):
        """
        Merge another metric group spec into the current one.

        The `other` MetricGroupsSpec overwrites existing keys.
        """
        seen = []
        for key, _ in self.definitions.items():
            if key in other.definitions:
                self.definitions[key].merge(other.definitions[key])
            seen.append(key)
        for key, definition in other.definitions.items():
            if key not in seen:
                self.definitions[key] = definition


converter.register_structure_hook(
    MetricGroupsSpec, lambda obj, _type: MetricGroupsSpec.from_dict(obj)
)



import datetime as dt
import re
from textwrap import dedent

import pytest
import pytz
import toml
from cattrs.errors import ClassValidationError

from metric_config_parser.analysis import AnalysisSpec
from metric_config_parser.experiment import Experiment
from metric_config_parser.metric import AnalysisPeriod
from metric_config_parser.outcome import OutcomeSpec


class TestOutcomes:
    def test_outcomes(self):
        config_str = dedent(
            """
            friendly_name = "Test outcome"
            description = "Outcome for testing"
            default_metrics = ["spam", "organic_search_count"]

            [metrics.spam]
            data_source = "main"
            select_expression = "1"

            [metrics.spam.statistics.bootstrap_mean]
            num_samples = 10
            pre_treatments = ["remove_nulls"]

            [metrics.organic_search_count.statistics.bootstrap_mean]

            [metrics.ad_clicks.statistics.bootstrap_mean]

            [data_sources.eggs]
            from_expression = "england.camelot"
            client_id_column = "client_info.client_id"
            """
        )

        outcome_spec = OutcomeSpec.from_dict(toml.loads(config_str))
        assert "spam" in outcome_spec.metrics
        assert "organic_search_count" in outcome_spec.metrics
        assert "ad_clicks" in outcome_spec.metrics
        assert "eggs" in outcome_spec.data_sources.definitions

        default_metrics = [m.name for m in outcome_spec.default_metrics]
        assert "spam" in default_metrics
        assert "organic_search_count" in default_metrics
        assert "ad_clicks" not in default_metrics

    def test_invalid_default_metrics(self):
        config_str = dedent(
            """
            friendly_name = "Test outcome"
            description = "Outcome for testing"
            default_metrics = ["spam"]

            [metrics.ad_clicks.statistics.bootstrap_mean]
            """
        )

        with pytest.raises(
            ValueError,
            match=re.escape(
                "Default metric MetricReference(name='spam') is not defined in outcome."
            ),
        ):
            OutcomeSpec.from_dict(toml.loads(config_str))

    def test_resolving_outcomes(self, experiments, config_collection):
        config_str = dedent(
            """
            [metrics]
            weekly = ["view_about_logins", "my_cool_metric"]
            daily = ["my_cool_metric"]

            [metrics.my_cool_metric]
            data_source = "main"
            select_expression = "{{agg_histogram_mean('payload.content.my_cool_histogram')}}"
            friendly_name = "Cool metric"
            description = "Cool cool cool 😎"
            bigger_is_better = false

            [metrics.my_cool_metric.statistics.bootstrap_mean]

            [metrics.view_about_logins.statistics.bootstrap_mean]
            """
        )

        spec = AnalysisSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiments[5], config_collection)
        weekly_metrics = [s.metric.name for s in cfg.metrics[AnalysisPeriod.WEEK]]

        assert "view_about_logins" in weekly_metrics
        assert "my_cool_metric" in weekly_metrics

    def test_resolving_parameters(self, experiments, config_collection):
        config_str = dedent(
            """
            [parameters.id]
            value = "123"

            [metrics]
            weekly = ["view_about_logins", "my_cool_metric"]
            daily = ["my_cool_metric"]

            [metrics.my_cool_metric]
            data_source = "main"
            select_expression = "{{agg_histogram_mean('payload.content.my_cool_histogram')}}"
            friendly_name = "Cool metric"
            description = "Cool cool cool 😎"
            bigger_is_better = false

            [metrics.my_cool_metric.statistics.bootstrap_mean]

            [metrics.view_about_logins.statistics.bootstrap_mean]
            """
        )

        spec = AnalysisSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiments[6], config_collection)
        weekly_metrics = [s.metric.name for s in cfg.metrics[AnalysisPeriod.WEEK]]

        assert len(weekly_metrics) == 3

        assert "id" in spec.parameters.definitions
        assert spec.parameters.definitions["id"].default == "700"
        assert spec.parameters.definitions["id"].value == "123"

        assert "view_about_logins" in weekly_metrics
        assert "my_cool_metric" in weekly_metrics

        outcome_metric = next(
            m.metric for m in cfg.metrics[AnalysisPeriod.WEEK] if m.metric.name == "sample_id_count"
        )
        assert outcome_metric.select_expression == "COUNTIF(sample_id = 123)"

    def test_resolving_parameters_distinct_by_branch(self, experiments, config_collection):
        config_str = dedent(
            """
            [parameters.id]
            distinct_by_branch = true

            value.branch_1 = "123"
            value.branch_2 = "456"

            default.branch_3 = "444"
            default.branch_1 = "444"

            [metrics]
            weekly = ["view_about_logins", "my_cool_metric"]
            daily = ["my_cool_metric"]

            [metrics.my_cool_metric]
            data_source = "main"
            select_expression = "{{agg_histogram_mean('payload.content.my_cool_histogram')}}"
            friendly_name = "Cool metric"
            description = "Cool cool cool 😎"
            bigger_is_better = false

            [metrics.my_cool_metric.statistics.bootstrap_mean]

            [metrics.view_about_logins.statistics.bootstrap_mean]
            """
        )

        spec = AnalysisSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiments[6], config_collection)
        weekly_metrics = [s.metric.name for s in cfg.metrics[AnalysisPeriod.WEEK]]

        assert len(weekly_metrics) == 3

        assert "id" in spec.parameters.definitions
        assert spec.parameters.definitions["id"].value["branch_1"] == "123"
        assert spec.parameters.definitions["id"].value["branch_2"] == "456"
        assert spec.parameters.definitions["id"].value["branch_3"] == "444"

        assert "view_about_logins" in weekly_metrics
        assert "my_cool_metric" in weekly_metrics

        outcome_metric = next(
            m.metric for m in cfg.metrics[AnalysisPeriod.WEEK] if m.metric.name == "sample_id_count"
        )
        assert outcome_metric.select_expression == (
            """COUNTIF(sample_id = CASE e.branch """
            """WHEN "branch_3" THEN "444" WHEN "branch_1" """
            """THEN "123" WHEN "branch_2" THEN "456" END)"""
        )

    def test_resolving_parameters_default_value(self, experiments, config_collection):
        config_str = dedent(
            """
            [metrics]
            weekly = ["view_about_logins", "my_cool_metric"]
            daily = ["my_cool_metric"]

            [metrics.my_cool_metric]
            data_source = "main"
            select_expression = "{{agg_histogram_mean('payload.content.my_cool_histogram')}}"
            friendly_name = "Cool metric"
            description = "Cool cool cool 😎"
            bigger_is_better = false

            [metrics.my_cool_metric.statistics.bootstrap_mean]

            [metrics.view_about_logins.statistics.bootstrap_mean]
            """
        )

        spec = AnalysisSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiments[6], config_collection)
        weekly_metrics = [s.metric.name for s in cfg.metrics[AnalysisPeriod.WEEK]]

        assert len(weekly_metrics) == 3

        assert "id" in spec.parameters.definitions
        assert spec.parameters.definitions["id"].value == "700"
        assert spec.parameters.definitions["id"].default == "700"

        assert "view_about_logins" in weekly_metrics
        assert "my_cool_metric" in weekly_metrics

        outcome_metric = next(
            m.metric for m in cfg.metrics[AnalysisPeriod.WEEK] if m.metric.name == "sample_id_count"
        )
        assert outcome_metric.select_expression == "COUNTIF(sample_id = 700)"

    def test_resolving_parameters_default_value_distinct_by_branch(
        self, experiments, config_collection
    ):
        config_str = dedent(
            """

            [parameters.id]
            # value.branch_1 = ""  # this will use default value from outcome
            value.branch_2 = 2  # this will use default value from outcome
            distinct_by_branch = true

            [metrics]
            weekly = ["view_about_logins", "my_cool_metric"]
            daily = ["my_cool_metric"]

            [metrics.my_cool_metric]
            data_source = "main"
            select_expression = "{{agg_histogram_mean('payload.content.my_cool_histogram')}}"
            friendly_name = "Cool metric"
            description = "Cool cool cool 😎"
            bigger_is_better = false

            [metrics.my_cool_metric.statistics.bootstrap_mean]

            [metrics.view_about_logins.statistics.bootstrap_mean]
            """
        )

        spec = AnalysisSpec.from_dict(toml.loads(config_str))
        cfg = spec.resolve(experiments[7], config_collection)
        weekly_metrics = [s.metric.name for s in cfg.metrics[AnalysisPeriod.WEEK]]

        assert len(weekly_metrics) == 3

        assert "id" in spec.parameters.definitions
        assert isinstance(spec.parameters.definitions["id"].default, dict)

        assert len(spec.parameters.definitions["id"].default) == 1
        assert "branch_2" not in spec.parameters.definitions["id"].default
        assert spec.parameters.definitions["id"].default["branch_1"] == 1

        assert spec.parameters.definitions["id"].value["branch_1"] == 1
        assert spec.parameters.definitions["id"].value["branch_2"] == 2

        assert "view_about_logins" in weekly_metrics
        assert "my_cool_metric" in weekly_metrics

        outcome_metric = next(
            m.metric for m in cfg.metrics[AnalysisPeriod.WEEK] if m.metric.name == "sample_id_count"
        )
        assert outcome_metric.select_expression == (
            """COUNTIF(sample_id = CASE e.branch """
            """WHEN "branch_1" THEN "1" WHEN "branch_2" THEN "2" END)"""
        )

    def test_resolving_parameters_distinct_by_branch_missing_branch_name_raises(self):
        """
        If distinct_by_branch is set to `true`
        `id.branch_name` value should be specified
        otherwise we raise an exception
        """

        config_str = dedent(
            """
            [parameters.id]
            id.value = "1234"
            value = "567"
            distinct_by_branch = true

            [metrics]
            weekly = ["view_about_logins", "my_cool_metric"]
            daily = ["my_cool_metric"]

            [metrics.my_cool_metric]
            data_source = "main"
            select_expression = "{{agg_histogram_mean('payload.content.my_cool_histogram')}}"
            friendly_name = "Cool metric"
            description = "Cool cool cool 😎"
            bigger_is_better = false

            [metrics.my_cool_metric.statistics.bootstrap_mean]

            [metrics.view_about_logins.statistics.bootstrap_mean]
            """
        )

        with pytest.raises(ClassValidationError):
            AnalysisSpec.from_dict(toml.loads(config_str))

    def test_unsupported_platform_outcomes(self, config_collection):
        spec = AnalysisSpec.from_dict(toml.loads(""))
        experiment = Experiment(
            experimenter_slug="test_slug",
            type="pref",
            status="Live",
            start_date=dt.datetime(2019, 12, 1, tzinfo=pytz.utc),
            end_date=dt.datetime(2020, 3, 1, tzinfo=pytz.utc),
            proposed_enrollment=7,
            branches=[],
            normandy_slug="normandy-test-slug",
            reference_branch=None,
            is_high_population=True,
            outcomes=["performance"],
            app_name="fenix",
        )

        with pytest.raises(
            ValueError, match=re.escape("Outcome performance doesn't support the platform 'fenix'")
        ):
            spec.resolve(experiment, config_collection)



import copy
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Optional

import attr

from metric_config_parser.alert import Alert, AlertsSpec
from metric_config_parser.data_source import DataSourcesSpec
from metric_config_parser.dimension import Dimension, DimensionsSpec
from metric_config_parser.metric import MetricsSpec, Summary
from metric_config_parser.parameter import ParameterSpec
from metric_config_parser.project import ProjectConfiguration, ProjectSpec
from metric_config_parser.util import converter

if TYPE_CHECKING:
    from metric_config_parser.config import ConfigCollection
    from metric_config_parser.definition import DefinitionSpecSub
    from metric_config_parser.experiment import Experiment


@attr.s(auto_attribs=True)
class MonitoringConfiguration:
    """
    Represents configuration options.

    All references, for example to data sources, have been resolved in this representation.
    Instead of instantiating this directly, consider using MonitoringSpec.resolve().
    """

    project: ProjectConfiguration | None = None
    metrics: list[Summary] = attr.Factory(list)
    dimensions: list[Dimension] = attr.Factory(list)
    alerts: list[Alert] = attr.Factory(list)


@attr.s(auto_attribs=True)
class MonitoringSpec:
    """
    Represents a configuration file.

    The expected use is like:
        MonitoringSpec.from_dict(toml.load(my_configuration_file)).resolve()
    which will produce a fully populated, concrete `MonitoringConfiguration`.
    """

    metrics: MetricsSpec = attr.Factory(MetricsSpec)
    data_sources: DataSourcesSpec = attr.Factory(DataSourcesSpec)
    project: ProjectSpec = attr.Factory(ProjectSpec)
    dimensions: DimensionsSpec = attr.Factory(DimensionsSpec)
    alerts: AlertsSpec = attr.Factory(AlertsSpec)
    parameters: ParameterSpec = attr.Factory(ParameterSpec)
    _resolved: bool = False

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> "MonitoringSpec":
        """Create a `MonitoringSpec` from a dict."""
        d = {k.lower(): v for k, v in d.items()}
        return converter.structure(d, cls)

    @classmethod
    def from_definition_spec(
        cls,
        spec: "DefinitionSpecSub",
        project: Optional["ProjectSpec"] = None,
    ) -> "MonitoringSpec":
        from metric_config_parser.definition import DefinitionSpec

        if not isinstance(spec, MonitoringSpec) and not isinstance(spec, DefinitionSpec):
            raise ValueError(f"Cannot create MonitoringSpec from {spec}")

        if project is None:
            if isinstance(spec, MonitoringSpec):
                return cls(
                    metrics=spec.metrics,
                    data_sources=spec.data_sources,
                    dimensions=spec.dimensions,
                    alerts=spec.alerts,
                    parameters=spec.parameters,
                    project=spec.project,
                )
            else:
                return cls(
                    metrics=spec.metrics,
                    data_sources=spec.data_sources,
                    dimensions=spec.dimensions,
                    parameters=spec.parameters,
                )
        else:
            return cls(
                metrics=spec.metrics,
                data_sources=spec.data_sources,
                dimensions=spec.dimensions,
                alerts=spec.alerts,
                project=project,
                parameters=spec.parameters,
            )

    def resolve(
        self, experiment: Optional["Experiment"], configs: "ConfigCollection"
    ) -> MonitoringConfiguration:
        """Create a `MonitoringConfiguration` from the spec."""
        if self._resolved:
            raise Exception("Can't resolve an MonitoringSpec twice")

        self._resolved = True

        resolved_project = self.project.resolve(self, experiment, configs)

        # filter to only have metrics that actually need to be monitored
        metrics = []
        for metric_ref in {p.name for p in self.project.metrics}:
            if metric_ref in self.metrics.definitions:
                metrics += self.metrics.definitions[metric_ref].resolve(
                    self, resolved_project, configs
                )
            else:
                raise ValueError(f"No definition for metric {metric_ref}.")

        # filter to only have dimensions that actually are in use
        dimensions = []
        for dimension_ref in {d.name for d in self.project.population.dimensions}:
            if dimension_ref in self.dimensions.definitions:
                dimensions.append(
                    self.dimensions.definitions[dimension_ref].resolve(
                        self, resolved_project, configs
                    )
                )
            else:
                raise ValueError(f"No definition for dimension {dimension_ref}.")

        # filter to only have alerts that actually are in use
        alerts = []
        for alert_ref in {d.name for d in self.project.alerts}:
            if alert_ref in self.alerts.definitions:
                alerts.append(
                    self.alerts.definitions[alert_ref].resolve(self, resolved_project, configs)
                )
            else:
                raise ValueError(f"No definition for alert {alert_ref}.")

        return MonitoringConfiguration(
            project=resolved_project,
            metrics=metrics,
            dimensions=dimensions,
            alerts=alerts,
        )

    def merge(self, other: Optional["DefinitionSpecSub"]):
        """Merge another monitoring spec into the current one."""
        from metric_config_parser.definition import DefinitionSpec

        if other:
            if isinstance(other, MonitoringSpec):
                self.project.merge(other.project)
            if isinstance(other, MonitoringSpec | DefinitionSpec):
                self.dimensions.merge(other.dimensions)
            if isinstance(other, MonitoringSpec | DefinitionSpec):
                self.alerts.merge(other.alerts)
            self.data_sources.merge(other.data_sources)
            self.metrics.merge(other.metrics)

    @classmethod
    def default_for_platform_or_type(
        cls, platform: str, configs: "ConfigCollection"
    ) -> "MonitoringSpec":
        """Return the default config for the provided platform."""
        default_metrics = configs.get_platform_defaults(platform)

        if default_metrics is None or not hasattr(default_metrics, "project"):
            spec = cls()
        else:
            spec = cls.from_definition_spec(default_metrics)

        return copy.deepcopy(spec)



import pytest

from metric_config_parser.metric import AnalysisPeriod, MetricDefinition
from metric_config_parser.parameter import ParameterDefinition


class TestMetricDefinition:
    @pytest.mark.parametrize(
        ("input", "expected"),
        [
            (
                [
                    {"param": ParameterDefinition(name="param", value="1")},
                    "param = {{ parameters.param }}",
                ],
                "param = 1",
            ),
            (
                [
                    {"param": ParameterDefinition(name="param", value="1")},
                    "{{ parameters.param }}",
                ],
                "1",
            ),
            ([{"param": ParameterDefinition(name="param", value="1")}, ""], ""),
            (
                [
                    {
                        "param": ParameterDefinition(
                            name="param",
                            distinct_by_branch=True,
                            value={"branch_1": "1"},
                        )
                    },
                    "",
                ],
                "",
            ),
            (
                [
                    {
                        "param": ParameterDefinition(
                            name="param",
                            distinct_by_branch=True,
                            value={"branch_1": "1"},
                        )
                    },
                    "{{parameters.param}}",
                ],
                'CASE e.branch WHEN "branch_1" THEN "1" END',
            ),
            (
                [
                    {
                        "param": ParameterDefinition(
                            name="param",
                            distinct_by_branch=True,
                            value={
                                "branch_1": "1",
                                "branch_2": "2",
                            },
                        )
                    },
                    "COUNTIF(id = {{parameters.param}})",
                ],
                (
                    """COUNTIF(id = CASE e.branch """
                    """WHEN "branch_1" THEN "1" WHEN "branch_2" THEN "2" END)"""
                ),
            ),
        ],
    )
    def test_generate_select_expression(self, input, expected, config_collection):
        """
        In case ParameterDefinition object is passed we just return the value,
        if List[ParameterDefinition] is given then we need to generate the whole select statement
        """

        param_definition, select_template = input

        actual = MetricDefinition.generate_select_expression(
            param_definition, select_template, config_collection
        )
        assert expected == actual

    def test_analysis_periods_conflicts(self):
        for test_period in AnalysisPeriod:
            for period in [p for p in AnalysisPeriod if p != test_period]:
                assert not period.value.startswith(f"{test_period.value}_")



import copy
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Optional

import attr

if TYPE_CHECKING:
    from .config import ConfigCollection
    from .definition import DefinitionSpec, DefinitionSpecSub
    from .outcome import OutcomeSpec

from .data_source import DataSourcesSpec
from .experiment import Experiment, ExperimentConfiguration, ExperimentSpec
from .metric import MetricReference, MetricsConfigurationType, MetricsSpec
from .parameter import ParameterDefinition, ParameterSpec
from .segment import SegmentsSpec
from .util import converter


@attr.s(auto_attribs=True)
class AnalysisConfiguration:
    """A fully concrete representation of the configuration for an experiment.

    Instead of instantiating this directly, consider using AnalysisSpec.resolve().
    """

    experiment: ExperimentConfiguration
    metrics: MetricsConfigurationType


@attr.s(auto_attribs=True)
class AnalysisSpec:
    """Represents a configuration file.

    The expected use is like:
        AnalysisSpec.from_dict(toml.load(my_configuration_file)).resolve(an_experimenter_object)
    which will produce a fully populated, concrete AnalysisConfiguration.
    """

    metrics: MetricsSpec = attr.Factory(MetricsSpec)
    data_sources: DataSourcesSpec = attr.Factory(DataSourcesSpec)
    segments: SegmentsSpec = attr.Factory(SegmentsSpec)
    parameters: ParameterSpec = attr.Factory(ParameterSpec)
    experiment: ExperimentSpec = attr.Factory(ExperimentSpec)
    _resolved: bool = False

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> "AnalysisSpec":
        return converter.structure(d, cls)

    @classmethod
    def from_definition_spec(
        cls, spec: "DefinitionSpec", experiment: Optional["ExperimentSpec"] = None
    ) -> "AnalysisSpec":
        if experiment is None:
            return cls(
                metrics=spec.metrics,
                data_sources=spec.data_sources,
                segments=spec.segments,
                parameters=spec.parameters,
            )
        else:
            return cls(
                metrics=spec.metrics,
                data_sources=spec.data_sources,
                segments=spec.segments,
                parameters=spec.parameters,
                experiment=experiment,
            )

    @classmethod
    def default_for_experiment(
        cls, experiment: "Experiment", configs: "ConfigCollection"
    ) -> "AnalysisSpec":
        """Return the default spec based on the experiment type."""
        default_metrics = configs.get_platform_defaults(experiment.app_name)

        if default_metrics is None or not isinstance(default_metrics, AnalysisSpec):
            default_metrics = cls()

        type_metrics = configs.get_platform_defaults(experiment.type)

        if type_metrics is not None or not isinstance(default_metrics, AnalysisSpec):
            default_metrics.merge(type_metrics)

        return copy.deepcopy(default_metrics)

    def resolve(
        self,
        experiment: "Experiment",
        configs: "ConfigCollection",
    ) -> AnalysisConfiguration:
        if self._resolved:
            raise Exception("Can't resolve an AnalysisSpec twice")
        self._resolved = True

        for slug in experiment.outcomes:
            outcome = configs.spec_for_outcome(slug, experiment.app_name)

            if outcome is not None:
                self.merge_outcome(outcome)
            else:
                raise ValueError(
                    f"Outcome {slug} doesn't support the platform '{experiment.app_name}'"
                )

            self.merge_parameters(outcome.parameters)

        resolved_experiment = self.experiment.resolve(self, experiment, configs)
        metrics = self.metrics.resolve(self, resolved_experiment, configs)

        return AnalysisConfiguration(resolved_experiment, metrics)

    def merge(self, other: "DefinitionSpecSub"):
        """Merges another analysis spec into the current one."""
        if isinstance(other, AnalysisSpec):
            self.experiment.merge(other.experiment)
        self.metrics.merge(other.metrics)
        self.data_sources.merge(other.data_sources)
        if isinstance(other, AnalysisSpec) or hasattr(other, "segments"):
            self.segments.merge(other.segments)  # type: ignore

    def merge_outcome(self, other: "OutcomeSpec"):
        """Merges an outcome snippet into the analysis spec."""
        metrics = [MetricReference(metric_name) for metric_name, _ in other.metrics.items()]

        # metrics defined in outcome snippets are only computed for
        # weekly and overall analysis windows
        outcome_spec = MetricsSpec(
            daily=[],
            weekly=metrics,
            days28=[],
            overall=metrics,
            definitions=other.metrics,
        )
        outcome_spec.merge(self.metrics)
        other.data_sources.merge(self.data_sources)
        self.data_sources = other.data_sources
        self.metrics = outcome_spec

        if other.parameters:
            self.merge_parameters(other.parameters)

    @staticmethod
    def _merge_param(
        param_1: "ParameterDefinition", param_2: "ParameterDefinition"
    ) -> "ParameterDefinition":
        """
        Takes in two ParameterDefinitions and merges them together into
        a single ParameterDefinition.

        param_2 is used for setting default values if missing in param_1
        """

        default_value = param_1.default or param_2.default
        value = param_1.value or param_2.value or default_value

        final_value = (
            {**(default_value or {}), **value}
            if isinstance(value, dict) and isinstance(default_value, dict)
            else value
        )

        return ParameterDefinition(
            name=getattr(param_1, "name", None) or param_2.name,
            friendly_name=getattr(param_1, "friendly_name", None) or param_2.friendly_name,
            description=getattr(param_1, "description", None) or param_2.description,
            value=(dict(final_value.items()) if isinstance(final_value, dict) else final_value),
            default=getattr(param_1, "default", None)
            or default_value
            or ({} if isinstance(final_value, dict) else None),
            distinct_by_branch=getattr(param_1, "distinct_by_branch", None)
            or param_2.distinct_by_branch,
        ).validate()

    def merge_parameters(self, other: "ParameterSpec") -> None:
        """
        Merges Outcome parameters with external config parameters.

        'self.parameters' -> contains custom config defined parameters
        'other' -> contains outcome defined
        """

        for param in other.definitions:
            external_config_param_settings = self.parameters.definitions.get(
                param, ParameterDefinition(name=param)
            )

            self.parameters.definitions[param] = AnalysisSpec._merge_param(
                external_config_param_settings, other.definitions[param]
            )



import re
import shutil
import tempfile
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path

import cattr
import pytz

converter = cattr.Converter()


@contextmanager
def TemporaryDirectory():
    name = Path(tempfile.mkdtemp())
    try:
        yield name
    finally:
        if name.exists():
            shutil.rmtree(name)


def parse_date(yyyy_mm_dd: str | None) -> datetime | None:
    if not yyyy_mm_dd:
        return None
    return datetime.strptime(yyyy_mm_dd, "%Y-%m-%d").replace(tzinfo=pytz.utc)


def is_valid_slug(slug: str) -> bool:
    """Returns whether a slug name is valid."""
    return bool(re.match(r"^[a-zA-Z0-9_]+$", slug))



import enum
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional

import attr

from metric_config_parser.alert import AlertReference
from metric_config_parser.metric import MetricReference
from metric_config_parser.metric_group import MetricGroup, MetricGroupsSpec
from metric_config_parser.population import PopulationConfiguration, PopulationSpec
from metric_config_parser.util import converter, parse_date

if TYPE_CHECKING:
    from metric_config_parser.config import ConfigCollection
    from metric_config_parser.experiment import Experiment
    from metric_config_parser.monitoring import MonitoringSpec


class MonitoringPeriod(enum.Enum):
    """
    Monitoring period.

    Used as x-axis.
    """

    BUILD_ID = "build_id"
    DAY = "submission_date"


@attr.s(auto_attribs=True, kw_only=True)
class ProjectConfiguration:
    """Describes the interface for defining the project in configuration."""

    reference_branch: str = "control"
    app_name: str = "firefox_desktop"
    name: str | None = None
    xaxis: MonitoringPeriod = attr.ib(default=MonitoringPeriod.DAY)
    start_date: datetime | None = None
    end_date: datetime | None = None
    population: PopulationConfiguration = attr.Factory(PopulationConfiguration)
    compact_visualization: bool = False
    skip_default_metrics: bool = False
    skip: bool = False
    is_rollout: bool = False
    metric_groups: list[MetricGroup] = attr.Factory(list)


def _validate_yyyy_mm_dd(instance: Any, attribute: Any, value: Any) -> None:
    parse_date(value)


@attr.s(auto_attribs=True, kw_only=True)
class ProjectSpec:
    """Describes the interface for defining the project."""

    name: str | None = None
    platform: str | None = None
    xaxis: MonitoringPeriod | None = None
    start_date: str | None = attr.ib(default=None, validator=_validate_yyyy_mm_dd)
    end_date: str | None = attr.ib(default=None, validator=_validate_yyyy_mm_dd)
    metrics: list[MetricReference] = attr.Factory(list)
    alerts: list[AlertReference] = attr.Factory(list)
    reference_branch: str | None = None
    population: PopulationSpec = attr.Factory(PopulationSpec)
    compact_visualization: bool = False
    skip_default_metrics: bool = False
    skip: bool = False
    is_rollout: bool = False
    metric_groups: MetricGroupsSpec = attr.Factory(MetricGroupsSpec)

    @classmethod
    def from_dict(cls, d: dict) -> "ProjectSpec":
        """Create a new `ProjectSpec` from a dictionary."""
        d = {k.lower(): v for k, v in d.items()}
        return converter.structure(d, cls)

    def resolve(
        self,
        spec: "MonitoringSpec",
        experiment: Optional["Experiment"],
        configs: "ConfigCollection",
    ) -> ProjectConfiguration:
        """Create a `ProjectConfiguration` from the spec."""

        project_config = ProjectConfiguration(
            name=self.name or (experiment.normandy_slug if experiment else None),
            xaxis=self.xaxis or MonitoringPeriod.DAY,
            start_date=parse_date(
                self.start_date
                or (
                    experiment.start_date.strftime("%Y-%m-%d")
                    if experiment and experiment.start_date
                    else None
                )
            ),
            end_date=parse_date(
                self.end_date
                or (
                    experiment.end_date.strftime("%Y-%m-%d")
                    if experiment and experiment.end_date
                    else None
                )
            ),
            reference_branch=self.reference_branch
            or (
                experiment.reference_branch
                if experiment and experiment.reference_branch
                else "control"
            ),
            compact_visualization=self.compact_visualization,
            skip_default_metrics=self.skip_default_metrics,
            skip=self.skip,
            app_name=self.platform or "firefox_desktop",
            is_rollout=self.is_rollout,
            metric_groups=[],
        )
        project_config.population = self.population.resolve(spec, project_config, configs)

        metric_groups = []
        for group_ref in [d for _, d in self.metric_groups.definitions.items()]:
            metric_groups.append(group_ref.resolve(spec, project_config, configs))

        project_config.metric_groups = metric_groups

        return project_config

    def merge(self, other: "ProjectSpec") -> None:
        """
        Merge another project spec into the current one.

        The `other` ProjectSpec overwrites existing keys.
        """
        for key in attr.fields_dict(type(self)):
            if key == "population":
                self.population.merge(other.population)
            elif key == "metrics":
                self.metrics += other.metrics
            elif key == "alerts":
                self.alerts += other.alerts
            elif key == "metric_groups":
                self.metric_groups.merge(other.metric_groups)
            else:
                setattr(self, key, getattr(other, key) or getattr(self, key))



class ConfigException(Exception):
    """Exception thrown when an experiment is invalid."""

    def __init__(self, message):
        super().__init__(message)


class InvalidConfigurationException(Exception):
    """Exception thrown when experiment configuration is invalid."""

    def __init__(self, message):
        super().__init__(message)


class DefinitionNotFound(Exception):
    """Exception thrown when a definition for a provided reference does not exist."""

    def __init__(self, message):
        super().__init__(message)


class UnexpectedKeyConfigurationException(InvalidConfigurationException):
    pass


class NoStartDateException(Exception):
    def __init__(self, normandy_slug, message="Experiment has no start date."):
        super().__init__(f"{normandy_slug} -> {message}")


class NoEndDateException(Exception):
    def __init__(self, normandy_slug, message="Experiment has no end date."):
        super().__init__(f"{normandy_slug} -> {message}")



import datetime as dt
import shutil
import tempfile
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

import attr
import jinja2
import toml
from git import Repo
from git.exc import InvalidGitRepositoryError
from jinja2 import StrictUndefined
from pytz import UTC

from metric_config_parser.data_source import DataSourceDefinition
from metric_config_parser.definition import DefinitionSpec, DefinitionSpecSub
from metric_config_parser.function import FunctionsSpec
from metric_config_parser.monitoring import MonitoringSpec
from metric_config_parser.segment import SegmentDataSourceDefinition, SegmentDefinition

from .analysis import AnalysisSpec
from .errors import UnexpectedKeyConfigurationException
from .experiment import Channel, Experiment
from .metric import MetricDefinition
from .outcome import OutcomeSpec
from .sql import generate_data_source_sql, generate_metrics_sql
from .util import TemporaryDirectory

if TYPE_CHECKING:
    from git.objects.commit import Commit

OUTCOMES_DIR = "outcomes"
DEFAULTS_DIR = "defaults"
DEFINITIONS_DIR = "definitions"
FUNCTIONS_FILE = "functions.toml"
JETSTREAM_CONFIG_URL = "https://github.com/mozilla/jetstream-config"


@attr.s(auto_attribs=True)
class Config:
    """Represent an external config file."""

    slug: str
    spec: DefinitionSpecSub
    last_modified: dt.datetime
    is_private: bool = False

    def validate(self, configs: "ConfigCollection", experiment: Experiment) -> None:
        if isinstance(self.spec, AnalysisSpec):
            analysis_spec = AnalysisSpec.default_for_experiment(experiment, configs)
            analysis_spec.merge(self.spec)
            resolved = analysis_spec.resolve(experiment, configs)
            # private configs need to override the default dataset
            if self.is_private and resolved.experiment.dataset_id is None:
                raise ValueError("dataset_id needs to be explicitly set for private experiments")
        elif isinstance(self.spec, MonitoringSpec):
            platform = (
                experiment.app_name if experiment else self.spec.project.platform
            ) or "firefox_desktop"
            monitoring_spec = MonitoringSpec.default_for_platform_or_type(
                platform,
                configs,
            )
            if experiment and experiment.is_rollout and platform == "firefox_desktop":
                rollout_spec = MonitoringSpec.default_for_platform_or_type("rollout", configs)
                monitoring_spec.merge(rollout_spec)
            monitoring_spec.merge(self.spec)
            monitoring_spec.resolve(experiment, configs)


def validate_config_settings(config_file: Path) -> None:
    """
    Implemented to resolve a Github issue:
        - https://github.com/mozilla/jetstream/issues/843

    Loads external config file and runs a number of validation steps on it:
    - Checks that all config keys/settings are lowercase
    - Checks for missing core config keys
    - Checks for unexpected core configuration keys
    - Checks that all segments defined under experiment have configuration in segments section
    - Checks if metric with custom config is defined in metrics.weekly or metrics.overall fields

    Returns None, if issues found with the configuration an Exception is raised
    """

    config = toml.loads(config_file.read_text())

    optional_core_config_keys = (
        "project",
        "population",
        "metrics",
        "experiment",
        "segments",
        "data_sources",
        "friendly_name",
        "description",
        "parameters",
        "alerts",
        "dimensions",
        "functions",
    )

    core_config_keys_specified = config.keys()

    # checks for unexpected core configuration keys
    unexpected_config_keys = set(core_config_keys_specified) - set(optional_core_config_keys)
    if unexpected_config_keys:
        err_msg = (
            f"Unexpected config key[s] found: {unexpected_config_keys}. "
            f"config_file: {str(config_file).split('/')[-1]}"
        )
        raise UnexpectedKeyConfigurationException(err_msg)

    return None


@attr.s(auto_attribs=True)
class DefaultConfig(Config):
    """
    Represents an external config files with platform-specific defaults.

    These config files are not associated to a specific experiment, since
    they are applied to all experiments.
    """

    def validate(self, configs: "ConfigCollection", _experiment: Experiment = None) -> None:
        dummy_experiment = Experiment(
            experimenter_slug="dummy-experiment",
            normandy_slug="dummy_experiment",
            type="v6",
            status="Live",
            branches=[],
            end_date=None,
            reference_branch="control",
            is_high_population=False,
            start_date=dt.datetime.now(UTC),
            proposed_enrollment=14,
            app_name=self.slug,
            channel=Channel.NIGHTLY,
        )
        spec = AnalysisSpec.default_for_experiment(dummy_experiment, configs)
        spec.merge(self.spec)
        spec.resolve(dummy_experiment, configs)


@attr.s(auto_attribs=True)
class Outcome:
    """Represents an external outcome snippet."""

    slug: str
    spec: OutcomeSpec
    platform: str
    commit_hash: str | None
    is_private: bool = False

    def validate(self, configs: "ConfigCollection") -> None:
        dummy_experiment = Experiment(
            experimenter_slug="dummy-experiment",
            normandy_slug="dummy_experiment",
            type="v6",
            status="Live",
            branches=[],
            end_date=None,
            reference_branch="control",
            is_high_population=False,
            start_date=dt.datetime.now(UTC),
            proposed_enrollment=14,
            app_name=self.platform,
            channel=Channel.NIGHTLY,
        )

        spec = AnalysisSpec.default_for_experiment(dummy_experiment, configs)
        spec.merge_outcome(self.spec)
        spec.merge_parameters(self.spec.parameters)
        spec.resolve(dummy_experiment, configs)


@attr.s(auto_attribs=True)
class DefinitionConfig(Config):
    """
    Represents an definition config file with definition that can be referenced in other configs.
    """

    platform: str = "firefox_desktop"

    @property
    def app_name(self) -> str:
        return self.platform

    def validate(self, configs: "ConfigCollection", _experiment: Experiment = None) -> None:
        dummy_experiment = Experiment(
            experimenter_slug="dummy-experiment",
            normandy_slug="dummy_experiment",
            type="v6",
            status="Live",
            branches=[],
            end_date=None,
            reference_branch="control",
            is_high_population=False,
            start_date=dt.datetime.now(UTC),
            proposed_enrollment=14,
            app_name=self.platform,
            channel=Channel.NIGHTLY,
        )

        if not isinstance(self.spec, DefinitionSpec):
            # this should not happen
            raise ValueError("Incorrect result type when parsing definition config")

        analysis_spec = AnalysisSpec.from_definition_spec(self.spec)
        analysis_spec.resolve(dummy_experiment, configs)


def entity_from_path(
    path: Path, is_private: bool = False
) -> Config | Outcome | DefaultConfig | DefinitionConfig | FunctionsSpec:
    is_outcome = path.parent.parent.name == OUTCOMES_DIR
    is_default_config = path.parent.name == DEFAULTS_DIR
    is_definition_config = path.parent.name == DEFINITIONS_DIR
    slug = path.stem

    validate_config_settings(path)

    config_dict = toml.loads(path.read_text())

    if is_outcome:
        platform = path.parent.name
        spec = OutcomeSpec.from_dict(config_dict)
        return Outcome(
            slug=slug,
            spec=spec,
            platform=platform,
            commit_hash=None,
            is_private=is_private,
        )
    elif is_default_config:
        if "project" in config_dict:
            # config is from opmon
            return DefaultConfig(
                slug=slug,
                spec=MonitoringSpec.from_dict(config_dict),
                last_modified=dt.datetime.fromtimestamp(path.stat().st_mtime, UTC),
                is_private=is_private,
            )
        else:
            return DefaultConfig(
                slug=slug,
                spec=AnalysisSpec.from_dict(config_dict),
                last_modified=dt.datetime.fromtimestamp(path.stat().st_mtime, UTC),
                is_private=is_private,
            )
    elif is_definition_config:
        if path.name == FUNCTIONS_FILE:
            return FunctionsSpec.from_dict(config_dict)
        else:
            return DefinitionConfig(
                slug=slug,
                spec=DefinitionSpec.from_dict(config_dict),
                last_modified=dt.datetime.fromtimestamp(path.stat().st_mtime, UTC),
                platform=slug,
                is_private=is_private,
            )

    if "project" in config_dict:
        # config is from opmon
        return Config(
            slug=slug,
            spec=MonitoringSpec.from_dict(config_dict),
            last_modified=dt.datetime.fromtimestamp(path.stat().st_mtime, UTC),
            is_private=is_private,
        )
    else:
        return Config(
            slug=slug,
            spec=AnalysisSpec.from_dict(config_dict),
            last_modified=dt.datetime.fromtimestamp(path.stat().st_mtime, UTC),
            is_private=is_private,
        )


@attr.s(auto_attribs=True)
class Repository:
    """Local repository config files are loaded from."""

    path: Path
    repo: Repo
    main_branch: str
    # indicates whether repository lives in a temporary directory that should be removed on del
    is_tmp_repo: bool = False
    commit_hash: str = "HEAD"

    def __del__(self):
        # remove the temporary directories repos have been saved in
        if self.is_tmp_repo:
            git_dir = Path(self.repo.git_dir)
            # don't delete repos that come from local directories
            if git_dir.parent.exists():
                shutil.rmtree(git_dir.parent)


@attr.s(auto_attribs=True)
class ConfigCollection:
    """
    Collection of experiment-specific configurations pulled in
    from an external GitHub repository.
    """

    configs: list[Config] = attr.Factory(list)
    outcomes: list[Outcome] = attr.Factory(list)
    defaults: list[DefaultConfig] = attr.Factory(list)
    definitions: list[DefinitionConfig] = attr.Factory(list)
    functions: FunctionsSpec | None = None
    repos: list[Repository] = attr.Factory(list)  # repos configs were loaded from
    is_private: bool = False

    repo_url = "https://github.com/mozilla/metric-hub"

    @classmethod
    def from_github_repo(
        cls,
        repo_url: str | None = None,
        is_private: bool = False,
        path: str | None = None,
        depth: int | None = None,
    ) -> "ConfigCollection":
        """Pull in external config files."""
        # download files to a persisted tmp directory
        tmp_dir = None
        is_tmp_repo = False
        if repo_url is not None and Path(repo_url).exists() and Path(repo_url).is_dir():
            if path:
                repo = Repo(repo_url)
            else:
                repo_path = Path(repo_url)
                dir_path = Path()
                depth = depth or 3

                # check if parent folder is a repository until search depth is exceeded
                while depth > 0:
                    try:
                        repo = Repo(repo_path)
                    except InvalidGitRepositoryError:
                        dir_path = Path(repo_path.name) / dir_path
                        repo_path = repo_path.parent
                        depth -= 1
                    else:
                        depth = 0

                repo = Repo(repo_path)
                path = str(dir_path)
                repo_url = str(repo_path)

            tmp_dir = Path(repo_url)
        else:
            tmp_dir = Path(tempfile.mkdtemp())
            is_tmp_repo = True
            if repo_url is not None and "/tree/" in repo_url:
                if not repo_url.endswith("/"):
                    repo_url += "/"
                repo_url, tree = repo_url.split("/tree/")
                branch, path = tree.split("/", 1)
                repo = Repo.clone_from(repo_url or cls.repo_url, tmp_dir)
                repo.git.checkout(branch)
            else:
                repo = Repo.clone_from(repo_url or cls.repo_url, tmp_dir)

        return ConfigCollection.from_local_repo(
            repo=repo,
            path=path,
            is_private=is_private,
            main_branch=repo.active_branch.name,
            is_tmp_repo=is_tmp_repo,
        )

    @classmethod
    def from_github_repos(
        cls, repo_urls: list[str] | None = None, is_private: bool = False
    ) -> "ConfigCollection":
        """Load configs from the provided repos."""
        if repo_urls is None or len(repo_urls) < 1:
            return ConfigCollection.from_github_repo()

        configs = None

        for repo in repo_urls:
            if configs is None:
                configs = ConfigCollection.from_github_repo(repo, is_private=is_private)
            else:
                collection = ConfigCollection.from_github_repo(repo, is_private=is_private)
                configs.merge(collection)
        return configs or ConfigCollection.from_github_repo()

    @classmethod
    def from_local_repo(
        cls, repo, path, is_private, main_branch, is_tmp_repo=False
    ) -> "ConfigCollection":
        """Load configs from a local repository."""

        files_path = Path(repo.git_dir).parent / path if path else Path(repo.git_dir).parent

        external_configs = []
        for config_file in files_path.glob("*.toml"):
            last_modified = next(repo.iter_commits("HEAD", paths=config_file)).committed_date
            config_json = toml.load(config_file)

            if "project" in config_json:
                # opmon spec
                spec: DefinitionSpecSub = MonitoringSpec.from_dict(config_json)
            else:
                spec = AnalysisSpec.from_dict(config_json)
                spec.experiment.is_private = spec.experiment.is_private or is_private

            external_configs.append(
                Config(
                    config_file.stem,
                    spec,
                    UTC.localize(dt.datetime.utcfromtimestamp(last_modified)),
                    is_private=is_private,
                )
            )

        outcomes = []
        for outcome_file in files_path.glob(f"{OUTCOMES_DIR}/*/*.toml"):
            commit_hash = next(repo.iter_commits("HEAD", paths=outcome_file)).hexsha

            outcomes.append(
                Outcome(
                    slug=outcome_file.stem,
                    spec=OutcomeSpec.from_dict(toml.load(outcome_file)),
                    platform=outcome_file.parent.name,
                    commit_hash=commit_hash,
                    is_private=is_private,
                )
            )

        default_configs = []
        for default_config_file in files_path.glob(f"{DEFAULTS_DIR}/*.toml"):
            last_modified = next(
                repo.iter_commits("HEAD", paths=default_config_file)
            ).committed_date

            default_config_json = toml.load(default_config_file)

            if "project" in default_config_json:
                # opmon spec
                spec = MonitoringSpec.from_dict(default_config_json)
            else:
                spec = AnalysisSpec.from_dict(default_config_json)
                spec.experiment.is_private = spec.experiment.is_private or is_private

            default_configs.append(
                DefaultConfig(
                    default_config_file.stem,
                    spec,
                    UTC.localize(dt.datetime.utcfromtimestamp(last_modified)),
                    is_private=is_private,
                )
            )

        definitions = []
        for definitions_config_file in files_path.glob(f"{DEFINITIONS_DIR}/*.toml"):
            last_modified = next(
                repo.iter_commits("HEAD", paths=definitions_config_file)
            ).committed_date

            definitions.append(
                DefinitionConfig(
                    definitions_config_file.stem,
                    DefinitionSpec.from_dict(toml.load(definitions_config_file)),
                    UTC.localize(dt.datetime.utcfromtimestamp(last_modified)),
                    platform=definitions_config_file.stem,
                    is_private=is_private,
                )
            )

        functions_spec = None
        for functions_file in files_path.glob(f"{DEFINITIONS_DIR}/{FUNCTIONS_FILE}"):
            functions_spec = FunctionsSpec.from_dict(toml.load(functions_file))

        return cls(
            external_configs,
            outcomes,
            default_configs,
            definitions,
            functions_spec,
            repos=[
                Repository(
                    repo=repo,
                    path=path,
                    main_branch=main_branch,
                    is_tmp_repo=is_tmp_repo,
                )
            ],
            is_private=is_private,
        )

    def as_of(self, timestamp: datetime) -> "ConfigCollection":
        """Get configs as they were at the provided timestamp."""
        if timestamp is None:
            return self

        config_collection = None

        # configs can be loaded from multiple different repos
        for repo in self.repos:
            # copy the original repo to a temporary path were we can go back in history
            with TemporaryDirectory() as tmp_dir:
                git_dir = Path(repo.repo.git_dir)
                # copy over repo content, except for lib/*, which doesn't contain metric information
                shutil.copytree(
                    git_dir.parent,
                    tmp_dir,
                    dirs_exist_ok=True,
                    ignore=shutil.ignore_patterns("lib*"),
                )
                tmp_repo = Repo(tmp_dir)

                # find the commit that got added just before the `timestamp`
                rev = "HEAD"  # use the HEAD commit by default if no other commits exist

                # make sure the most recent commit is checked out by checking out the main branch
                tmp_repo.git.checkout(repo.main_branch)

                # keep track of more recent commits to go back to in case invalid configs
                # got checked into main that cannot be parsed
                newer_commits: list[Commit] = []

                # start iterating through all commits starting at HEAD
                for commit in tmp_repo.iter_commits("HEAD"):
                    # check if the commit timestamp is older than the reference timestamp
                    if commit.committed_datetime <= timestamp:
                        rev = commit.hexsha
                        break

                    newer_commits.insert(0, commit)

                if commit:
                    # if there is no commit that is older than the reference timestamp,
                    # use the oldest commit that we could find (= last commit)
                    rev = commit.hexsha

                # checkout that commit
                tmp_repo.git.checkout(rev)

                # load configs as they were at the time of the commit
                try:
                    configs = self.from_local_repo(
                        tmp_repo,
                        repo.path,
                        self.is_private,
                        repo.main_branch,
                        is_tmp_repo=True,
                    )
                except Exception as e:
                    could_load_configs = False

                    # iterate through newer commits to find one that can be parsed
                    for newer_commit in newer_commits:
                        tmp_repo.git.checkout(newer_commit.hexsha)

                        try:
                            configs = self.from_local_repo(
                                tmp_repo,
                                repo.path,
                                self.is_private,
                                repo.main_branch,
                                is_tmp_repo=True,
                            )
                            could_load_configs = True
                            rev = newer_commit.hexsha
                            break
                        except Exception:
                            # continue searching
                            pass

                    if not could_load_configs:
                        # there is no newer commit, current state is broken
                        raise e

                repo.commit_hash = rev
                configs.repos = [repo]  # point to the original repo, instead of the temporary one

                if config_collection is None:
                    config_collection = configs
                else:
                    config_collection.merge(configs)

        if config_collection is None:
            return self

        return config_collection

    def spec_for_outcome(self, slug: str, platform: str) -> OutcomeSpec | None:
        """Return the spec for a specific outcome"""
        for outcome in self.outcomes:
            if outcome.slug == slug and outcome.platform == platform:
                return outcome.spec

        return None

    def spec_for_experiment(self, slug: str) -> AnalysisSpec | None:
        """Return the spec for a specific experiment."""
        for config in self.configs:
            if config.slug == slug and isinstance(config.spec, AnalysisSpec):
                return config.spec

        return None

    def spec_for_project(self, slug: str) -> MonitoringSpec | None:
        """Return the spec for a specific project."""
        for config in self.configs:
            if config.slug == slug and isinstance(config.spec, MonitoringSpec):
                return config.spec

        return None

    def get_platform_defaults(self, platform: str) -> DefinitionSpecSub | None:
        for default in self.defaults:
            if platform == default.slug:
                return default.spec

        return None

    def get_platform_definitions(self, platform: str) -> DefinitionSpecSub | None:
        for definition in self.definitions:
            if platform == definition.slug:
                return definition.spec

        return None

    def get_metric_definition(self, slug: str, app_name: str) -> MetricDefinition | None:
        for definition in self.definitions:
            if app_name == definition.platform:
                for metric_slug, metric in definition.spec.metrics.definitions.items():
                    if metric_slug == slug:
                        return metric

        return None

    def get_data_source_definition(self, slug: str, app_name: str) -> DataSourceDefinition | None:
        for definition in self.definitions:
            if app_name == definition.platform:
                for (
                    data_source_slug,
                    data_source,
                ) in definition.spec.data_sources.definitions.items():
                    if data_source_slug == slug:
                        return data_source
        return None

    def get_segment_data_source_definition(
        self, slug: str, app_name: str
    ) -> SegmentDataSourceDefinition | None:
        for definition in self.definitions:
            if app_name == definition.platform and not isinstance(definition.spec, MonitoringSpec):
                for (
                    segment_source_slug,
                    segment_source,
                ) in definition.spec.segments.data_sources.items():
                    if segment_source_slug == slug:
                        return segment_source

        return None

    def get_segment_definition(self, slug: str, app_name: str) -> SegmentDefinition | None:
        for definition in self.definitions:
            if app_name == definition.platform and not isinstance(definition.spec, MonitoringSpec):
                for (
                    segment_slug,
                    segment,
                ) in definition.spec.segments.definitions.items():
                    if segment_slug == slug:
                        return segment

        return None

    def get_segments_for_app(self, app_name: str) -> list[SegmentDefinition] | None:
        segments: list[SegmentDefinition] = []
        for definition in self.definitions:
            if app_name == definition.platform and not isinstance(definition.spec, MonitoringSpec):
                segments.extend(definition.spec.segments.definitions.values())

        return segments

    def get_metrics_sql(
        self,
        metrics: list[str],
        platform: str,
        group_by: list[str] | None = None,
        where: str | None = None,
        group_by_client_id: bool = True,
        group_by_submission_date: bool = True,
    ) -> str:
        """Generate a SQL query for the specified metrics."""
        if group_by is None:
            group_by = []
        return generate_metrics_sql(
            self,
            metrics=metrics,
            platform=platform,
            group_by=group_by,
            where=where,
            group_by_client_id=group_by_client_id,
            group_by_submission_date=group_by_submission_date,
        )

    def get_data_source_sql(
        self,
        data_source: str,
        platform: str,
        where: str | None = None,
        select_fields: bool = True,
        ignore_joins: bool = False,
    ) -> str:
        return generate_data_source_sql(
            self,
            data_source=data_source,
            platform=platform,
            where=where,
            select_fields=select_fields,
            ignore_joins=ignore_joins,
        )

    def get_env(self) -> jinja2.Environment:
        """
        Create a Jinja2 environment that understands the SQL agg_* helpers in mozanalysis.metrics.

        Just a wrapper to avoid leaking temporary variables to the module scope."""
        env = jinja2.Environment(autoescape=False, undefined=StrictUndefined)
        if self.functions is not None:
            for slug, function in self.functions.functions.items():
                env.globals[slug] = function.definition

        return env

    def merge(self, other: "ConfigCollection"):
        """
        Merge this config collection with another.

        Configs in `other` will take precedence.
        """
        # merge configs
        other_configs = {config.slug: config for config in deepcopy(other.configs)}
        configs = {config.slug: config for config in deepcopy(self.configs)}

        for slug, config in other_configs.items():
            if slug not in configs:
                configs[slug] = config
            else:
                configs[slug].spec.merge(config.spec)

        self.configs = list(configs.values())

        # merge outcomes
        outcomes = deepcopy(other.outcomes)
        slugs = [outcome.slug for outcome in outcomes]
        for outcome in self.outcomes:
            if outcome.slug not in slugs:
                outcomes.append(outcome)

        self.outcomes = outcomes

        # merge definitions
        other_definitions = {
            definition.slug: definition for definition in deepcopy(other.definitions)
        }
        definitions = {
            definition.slug: deepcopy(definition) for definition in deepcopy(self.definitions)
        }

        for slug, definition in other_definitions.items():
            if slug not in definitions:
                definitions[slug] = deepcopy(definition)
            else:
                definitions[slug].spec.merge(definition.spec)

        self.definitions = list(definitions.values())

        # merge defaults
        other_defaults = {default.slug: default for default in deepcopy(other.defaults)}
        defaults = {default.slug: default for default in deepcopy(self.defaults)}

        for slug, default in other_defaults.items():
            if slug not in defaults:
                defaults[slug] = deepcopy(default)
            else:
                defaults[slug].spec.merge(default.spec)

        self.defaults = list(defaults.values())

        # merge functions
        functions = deepcopy(other.functions.functions) if other.functions else {}
        slugs = [slug for slug, _ in functions.items()]
        for slug, function in self.functions.functions.items() if self.functions else {}:
            if slug not in slugs:
                functions[slug] = function

        if self.functions is None:
            self.functions = FunctionsSpec(functions=functions)
        else:
            self.functions.functions = functions

        self.repos += other.repos


@attr.s(auto_attribs=True)
class LocalConfigCollection(ConfigCollection):
    """
    Collection of experiment-specific configurations pulled in
    from a local directory (structured like metric-hub).

    Unless you are Experimenter, you likely want to use ConfigCollection instead.
    """

    @classmethod
    def from_local_path(cls, path: str, is_private: bool = False) -> "ConfigCollection":
        """Load configs from a local non-repo copy of a metric-hub-like folder structure.

        Args:
            path (str): path to the configs. Looks for TOML files in the
                following locations (all of which are optional):
                    - . (root)      - creates Config
                    - outcomes      - creates Outcome
                    - defaults      - creates DefaultConfig
                    - definitions   - creates DefinitionConfig
            is_private (bool): whether the configs are private
        """

        files_path = Path(path)

        external_configs = []
        last_modified = dt.datetime.now()
        for config_file in files_path.glob("*.toml"):
            config_json = toml.load(config_file)

            if "project" in config_json:
                # opmon spec
                spec: DefinitionSpecSub = MonitoringSpec.from_dict(config_json)
            else:
                spec = AnalysisSpec.from_dict(config_json)
                spec.experiment.is_private = spec.experiment.is_private or is_private

            external_configs.append(
                Config(
                    config_file.stem,
                    spec,
                    last_modified,
                    is_private=is_private,
                )
            )

        outcomes = []
        for outcome_file in files_path.glob(f"{OUTCOMES_DIR}/*/*.toml"):
            outcomes.append(
                Outcome(
                    slug=outcome_file.stem,
                    spec=OutcomeSpec.from_dict(toml.load(outcome_file)),
                    platform=outcome_file.parent.name,
                    commit_hash=None,
                    is_private=is_private,
                )
            )

        default_configs = []
        for default_config_file in files_path.glob(f"{DEFAULTS_DIR}/*.toml"):
            default_config_json = toml.load(default_config_file)

            if "project" in default_config_json:
                # opmon spec
                spec = MonitoringSpec.from_dict(default_config_json)
            else:
                spec = AnalysisSpec.from_dict(default_config_json)
                spec.experiment.is_private = spec.experiment.is_private or is_private

            default_configs.append(
                DefaultConfig(
                    default_config_file.stem,
                    spec,
                    last_modified,
                    is_private=is_private,
                )
            )

        definitions = []
        for definitions_config_file in files_path.glob(f"{DEFINITIONS_DIR}/*.toml"):
            definitions.append(
                DefinitionConfig(
                    definitions_config_file.stem,
                    DefinitionSpec.from_dict(toml.load(definitions_config_file)),
                    last_modified,
                    platform=definitions_config_file.stem,
                    is_private=is_private,
                )
            )

        functions_spec = None
        for functions_file in files_path.glob(f"{DEFINITIONS_DIR}/{FUNCTIONS_FILE}"):
            functions_spec = FunctionsSpec.from_dict(toml.load(functions_file))

        return cls(
            external_configs,
            outcomes,
            default_configs,
            definitions,
            functions_spec,
            repos=[],
            is_private=is_private,
        )

    @classmethod
    def from_github_repo(
        cls,
        repo_url: str | None = None,
        is_private: bool = False,
        path: str | None = None,
        depth: int | None = None,
    ):
        raise NotImplementedError(
            "`from_github_repo` is not valid for non-repo-based LocalConfigCollection. "
            + "Use ConfigCollection for this feature."
        )

    @classmethod
    def from_github_repos(
        cls, repo_urls: list[str] | None = None, is_private: bool = False
    ) -> "ConfigCollection":
        raise NotImplementedError(
            "`from_github_repos` is not valid for non-repo-based LocalConfigCollection. "
            + "Use ConfigCollection for this feature."
        )

    @classmethod
    def from_local_repo(
        cls, repo, path, is_private, main_branch, is_tmp_repo=False
    ) -> "ConfigCollection":
        raise NotImplementedError(
            "`from_local_repo` is not valid for non-repo-based LocalConfigCollection. "
            + "Use ConfigCollection for this feature."
        )

    @classmethod
    def as_of(self, timestamp: datetime) -> "ConfigCollection":
        raise NotImplementedError(
            "`as_of` is not valid for non-repo-based LocalConfigCollection. "
            + "Use ConfigCollection for this feature."
        )



from collections.abc import Callable, Mapping
from functools import partial
from typing import Any

import attr


@attr.s(auto_attribs=True)
class Function:
    slug: str
    definition: Callable
    friendly_name: str | None = None
    description: str | None = None


@attr.s(auto_attribs=True)
class FunctionsSpec:
    functions: dict[str, Function]

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> "FunctionsSpec":
        return cls(
            {
                slug: Function(
                    slug=slug,
                    definition=(
                        partial(
                            lambda select_expr, definition: definition.format(
                                select_expr=select_expr
                            ),
                            definition=fun["definition"],
                        )
                    ),
                    friendly_name=(fun.get("friendly_name", None)),
                    description=fun.get("description", None),
                )
                for slug, fun in d["functions"].items()
            }
        )



import enum
from textwrap import dedent
from typing import TYPE_CHECKING, Any

import attr

if TYPE_CHECKING:
    from .analysis import AnalysisSpec
    from .config import ConfigCollection
    from .experiment import ExperimentConfiguration

from .data_source import DataSource, DataSourceReference
from .util import converter


class AnalysisWindow(enum.Enum):
    """
    Predefined timelimits that can be used for defining when exposures
    should be computed.
    """

    ANALYSIS_WINDOW_START = "analysis_window_start"
    ANALYSIS_WINDOW_END = "analysis_window_end"
    ENROLLMENT_START = "enrollment_start"
    ENROLLMENT_END = "enrollment_end"


WindowLimit = int | AnalysisWindow | None


def structure_window_limit(value: Any, _klass: type) -> WindowLimit:
    try:
        return AnalysisWindow(value)
    except Exception:
        return int(value)


converter.register_structure_hook(WindowLimit, structure_window_limit)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class ExposureSignal:
    """
    Jetstream exposure signal representation.

    Jetstream exposure signals are supersets of mozanalysis exposure signals
    with some additional metdata required for analysis.
    """

    name: str
    data_source: DataSource
    select_expression: str
    friendly_name: str
    description: str
    window_start: WindowLimit = attr.ib(None)
    window_end: WindowLimit = attr.ib(None)

    @window_end.validator
    @window_start.validator
    def validate_window(self, _attribute, value):
        if value is not None and not isinstance(value, int):
            AnalysisWindow(value)


@attr.s(auto_attribs=True)
class ExposureSignalDefinition:
    """Describes the interface for defining an exposure signal in configuration."""

    name: str
    data_source: DataSourceReference
    select_expression: str
    friendly_name: str
    description: str
    window_start: WindowLimit = None
    window_end: WindowLimit = None

    def resolve(
        self,
        spec: "AnalysisSpec",
        conf: "ExperimentConfiguration",
        configs: "ConfigCollection",
    ) -> ExposureSignal:
        return ExposureSignal(
            name=self.name,
            data_source=self.data_source.resolve(spec, conf=conf, configs=configs),
            select_expression=self.select_expression,
            friendly_name=(
                dedent(self.friendly_name) if self.friendly_name else self.friendly_name
            ),
            description=(dedent(self.description) if self.description else self.description),
            window_start=self.window_start,
            window_end=self.window_end,
        )
